/*
This is the main file for the bot which contains all of the AI,
spawning code, etc. The different functions are seperate by
comments telling what they're generally for and such to make
it easier to understand.
*/

#include "extdll.h"
#include "util.h"
#include "cbase.h"
#include "player.h"
#include "gamerules.h"

//=============================================================
//BOT CLASS
//=============================================================
class CBot : public CBasePlayer //Derive a bot class from CBasePlayer and do nothing but set isnetclient to false.
{
public:
	virtual BOOL IsNetClient( void ) { return FALSE; } //Bots should return FALSE for this, they can't receive NET messages.
};



//=============================================================
//GENERIC FUNCTION USED IN BOT AI
//=============================================================
float BotChangeYaw ( int yawSpeed, CBasePlayer *bot ) //Generic yaw angling function
{
	float		ideal, current, move, speed;

	if (!bot || !bot->pev || bot->pev->deadflag != DEAD_NO)
		return 0;

	current = UTIL_AngleMod( bot->pev->angles.y );
	ideal = bot->pev->ideal_yaw;
	if (current != ideal)
	{
		speed = (float)yawSpeed * gpGlobals->frametime * 10;
		move = ideal - current;

		if (ideal > current)
		{
			if (move >= 180)
				move = move - 360;
		}
		else
		{
			if (move <= -180)
				move = move + 360;
		}

		if (move > 0)
		{
			if (move > speed)
				move = speed;
		}
		else
		{
			if (move < -speed)
				move = -speed;
		}
		
		bot->pev->angles.y = UTIL_AngleMod (current + move);
	}
	else
		move = 0;

	return move;
}

//=============================================================
//ITEM FUNCTIONS FOR BOT
//=============================================================
CBasePlayer	*ItemByIndex( int playerIndex ) //Get an item by its index number
{
	CBasePlayer *pPlayer = NULL;
	entvars_t *pev;

	if ( playerIndex > 0 && playerIndex <= gpGlobals->maxEntities )
	{
		edict_t *pPlayerEdict = INDEXENT( playerIndex );
		if ( pPlayerEdict && pPlayerEdict->v.solid == SOLID_TRIGGER)
		{
			pev = &pPlayerEdict->v;
			if (!pev)
				return NULL;
			pPlayer = GetClassPtr((CBasePlayer *)pev);

			if (!pPlayer)
				return NULL;
		}
	}
	
	return pPlayer;
}

int ItemPrior(CBasePlayer *bot, CBasePlayer *item) { //Function to see if we want an item

	if (!bot || !bot->pev || bot->pev->deadflag != DEAD_NO)
		return 0; //We're dead

	if (!item || !item->pev)
		return 0; //Item isn't here

	//Here, add your own item checks and return 1 if the checks are passed.
	//Currently, the bot will not go after any items because there are no
	//valid checks.

	return 0;
}

void SearchForItem(CBasePlayer *bot) { //Function to find items.
	for ( int i = 1; i <= gpGlobals->maxEntities; i++ )
	{
		CBasePlayer *search = ItemByIndex( i );

		if ( search && bot->FVisible(search) && !(search->pev->effects & EF_NODRAW) && search->pev->origin.z < bot->pev->origin.z + 5 && ItemPrior(bot, search) == 1 )
		{ //If we passed all the checks for finding the item, then use it as our item.
			bot->item = search;
		}
	}
}

void GoAfterItem(CBasePlayer *bot) { //AI function to seek items.
	Vector aimorg, aimvec;

	if (bot->item == NULL)
		return;

	bot->pev->ideal_yaw = UTIL_VecToYaw( bot->item->pev->origin - bot->pev->origin );
	aimorg = bot->item->pev->origin;
	aimvec = UTIL_VecToAngles(bot->pev->origin - bot->item->pev->origin);
	bot->pev->angles.x = aimvec.x;
	bot->pev->yaw_speed = 360;
	BotChangeYaw(bot->pev->yaw_speed, bot);

	bot->fwd_val = CVAR_GET_FLOAT( "sv_maxspeed" ); //Player running speed

	if (!bot->FVisible(bot->item) || bot->item->pev->effects & EF_NODRAW)
		bot->item = NULL;
}



//=============================================================
//ENEMY FUNCTIONS FOR BOT
//=============================================================
CBaseEntity	*GetClientByIndex( int playerIndex )
{
	EHANDLE pPlayer;
	pPlayer = NULL;

	if ( playerIndex > 0 && playerIndex <= gpGlobals->maxClients )
	{
		edict_t *pPlayerEdict = INDEXENT( playerIndex );
		if ( pPlayerEdict && !pPlayerEdict->free && pPlayerEdict->v.deadflag == DEAD_NO && (pPlayerEdict->v.flags & FL_CLIENT) )
		{
			pPlayer = GetClassPtr((CBaseEntity *)&pPlayerEdict->v);
			if (pPlayer == NULL || !pPlayer->pev || !(pPlayer->pev->flags & FL_CLIENT))
				return NULL;
		}
	}
	
	return pPlayer;
}

void SearchForEnemy(CBasePlayer *bot) { //Function to find enemies.
	EHANDLE nearestenemy;
	nearestenemy = NULL;

	for ( int i = 1; i <= gpGlobals->maxClients; i++ )
	{
		EHANDLE pPlayer;
		pPlayer = GetClientByIndex(i);
		float Dist1;
		float closest1 = 9999999999;
		
		if (pPlayer != NULL && pPlayer->pev)
			Dist1 = (bot->pev->origin - pPlayer->pev->origin).Length();

		if ( pPlayer != NULL && pPlayer != bot && pPlayer->pev && pPlayer != bot && pPlayer->pev->deadflag == DEAD_NO && bot->FVisible(pPlayer))
		{
			if ((bot->pev->origin - pPlayer->pev->origin).Length() < closest1) {
				closest1 = (bot->pev->origin - pPlayer->pev->origin).Length();
				nearestenemy = pPlayer;
			}
		}
	} //This will find the nearest enemy

	if (nearestenemy != NULL) {
		bot->enemy = nearestenemy;
	} //If we found someone, set them as our enemy.
}

void GoAfterEnemy(CBasePlayer *bot) { //AI function to kill enemies.
	//You'll want to add code here to make the bot handle weapons
	//intelligently, select the best weapon, reload, manuever for
	//combat, etc.

	Vector aimorg, aimvec;

	if (bot->enemy == NULL)
		return;

	bot->pev->ideal_yaw = UTIL_VecToYaw( bot->enemy->pev->origin - bot->pev->origin );
	aimorg = bot->enemy->pev->origin;
	aimvec = UTIL_VecToAngles(bot->pev->origin - bot->enemy->pev->origin);
	bot->pev->angles.x = aimvec.x;
	bot->pev->yaw_speed = 360;
	BotChangeYaw(bot->pev->yaw_speed, bot);

	bot->fwd_val = CVAR_GET_FLOAT( "sv_maxspeed" );
	
	if (rand()%10 < 5)
		bot->pev->button = 0;
	else
		bot->pev->button = IN_ATTACK; //Run at the enemy while randomly
									  //hitting fire.
	if (!bot->FVisible(bot->enemy) || bot->enemy->pev->deadflag != DEAD_NO)
		bot->enemy = NULL;
}



//=============================================================
//ROAMING FUNCTION FOR BOT
//=============================================================
void RoamRandomly(CBasePlayer *bot) { //Very basic AI function to roam around.

/*
This function may look somewhat confusing to you if you're new
to this sort of thing, but don't worry about it. Just look over
it after you see how the bot behaves as a result of it and you'll
see how it's done and what does what eventually.
*/

	TraceResult tr;
	Vector		vecLookerOrigin;
	Vector		vecTargetOrigin;
	Vector		vecDownOrigin;

	UTIL_MakeVectors(bot->pev->angles);

	if (bot->timechangedir < gpGlobals->time) {
		bot->timechangedir = gpGlobals->time + 1;
		if (rand()%10 < 2) {
			if (bot->roamdir == 1)
				bot->roamdir = 0;
			else
				bot->roamdir = 1;
		}
	}

	vecLookerOrigin = bot->pev->origin + bot->pev->view_ofs;
	vecTargetOrigin = bot->pev->origin + gpGlobals->v_forward*100;
	vecDownOrigin = bot->pev->origin + gpGlobals->v_forward*39;
	vecDownOrigin.z -= 100;

	UTIL_TraceLine(vecLookerOrigin, vecTargetOrigin, ignore_monsters, ignore_glass, ENT(bot->pev)/*pentIgnore*/, &tr);
	
	if (tr.flFraction != 1.0)
	{
		if (bot->roamdir == 1)
			bot->pev->ideal_yaw += rand()%20;
		else
			bot->pev->ideal_yaw -= rand()%20;
	}
	vecLookerOrigin = bot->pev->origin;
	UTIL_TraceLine(vecLookerOrigin, vecDownOrigin, ignore_monsters, ignore_glass, ENT(bot->pev)/*pentIgnore*/, &tr);
	
	int tracedist = (vecLookerOrigin - tr.vecEndPos).Length();

	if (tracedist < 34 && tr.flFraction != 1.0) {
		if (bot->roamdir == 1)
			bot->pev->ideal_yaw += rand()%20;
		else
			bot->pev->ideal_yaw -= rand()%20;
		bot->fwd_val = 0;
		bot->side_val = 0;
	}
	else if (tr.flFraction == 1.0)
	{
		if (bot->roamdir == 1)
			bot->pev->ideal_yaw += rand()%20;
		else
			bot->pev->ideal_yaw -= rand()%20;
		bot->fwd_val = 0;
		bot->side_val = 0;
	}
	else {
		bot->pev->yaw_speed = 360;
		BotChangeYaw(bot->pev->yaw_speed, bot);
		bot->fwd_val = CVAR_GET_FLOAT( "sv_maxspeed" );
		bot->side_val = 0;
	}
}



//=============================================================
//MAIN BOT AI FUNCTIONS
//=============================================================
void BasicAI(CBasePlayer *bot) {
	if (bot->pev->health < 1 || bot->pev->deadflag != DEAD_NO) { //Dead
		if (rand()%10 < 5)
			bot->pev->button = 0;
		else
			bot->pev->button = IN_ATTACK;

		return;
	} //Randomly slap attack button if dead until we respawn.

	if (bot->item == NULL) { //If we aren't after an item, look for one.
		SearchForItem(bot);
	}
	
	if (bot->item != NULL) { //If we see an item, get it.
		GoAfterItem(bot);
		return;
	}

	if (bot->enemy == NULL && bot->item == NULL) { //If we can't find an item and
		SearchForEnemy(bot);					   //don't have an enemy, look for
	}											   //an enemy.
	
	if (bot->enemy != NULL && bot->item == NULL) { //If we found an enemy, go after it.
		GoAfterEnemy(bot);
		return;
	}

	if (bot->enemy == NULL && (bot->pev->button & IN_ATTACK))
		bot->pev->button -= IN_ATTACK; //Shouldn't fire if there's no enemy.

	if (bot->enemy == NULL && bot->item == NULL) { //Nothing is in sight that we're
		RoamRandomly(bot);						   //interested in, so roam randomly.
	}
}

//VALUES FOR CALCULATING MSEC VALUE
int msecnum;
float msecdel;
float msecval;

void CBasePlayer::BThink () {
	//The following code calculates the msec value for the physics function.
	//It needs to be dependant on framerate so the bot will move at the
	//correct speed.
	if (msecdel <= gpGlobals->time) {
		msecdel = gpGlobals->time + 1;
		if (msecnum > 0)
			msecval = 800/msecnum;
		msecnum = 0;
	}
	else
		msecnum++;

	if (msecval < 5)
		msecval = 5;

	if (msecval > 100)
		msecval = 100;
	//That's all a bit of a hack, but it works ok.


	BasicAI(this);

	pev->v_angle.y = pev->angles.y;	//A lot of our angle manipulation in the
	pev->v_angle.x = pev->angles.x;	//AI deals only with the angles value, and
									//we want our view angles to be the same
									//(but only side and vertical).

	g_engfuncs.pfnRunPlayerMove(edict(), pev->v_angle, fwd_val, side_val, 0, pev->button, 0, 50);
	//g_engfuncs.pfnRunPlayerMove(<bot edict>, <bot's view angles>, <forward move value>, <side move value> (strafing), <upward move value> (seems to do nothing), <buttons> (IN_ATTACK, etc.), <impulse number>, <msec value>);
}



//=============================================================
//BOT SPAWNING FUNCTIONS
//=============================================================
void BotClientPutInServer( edict_t *pEntity )
{
	CBot *pPlayer;

	entvars_t *pev = &pEntity->v;

	pPlayer = GetClassPtr((CBot *)pev); //Link bot object to the edict

	pPlayer->isabot = 1;

	pPlayer->Spawn() ;

	//Set the bot's user info:
	g_engfuncs.pfnSetClientKeyValue(pPlayer->entindex(), g_engfuncs.pfnGetInfoKeyBuffer(pEntity), "name", "Bot");
	g_engfuncs.pfnSetClientKeyValue(pPlayer->entindex(), g_engfuncs.pfnGetInfoKeyBuffer(pEntity), "model", "gordon");
	g_engfuncs.pfnSetClientKeyValue(pPlayer->entindex(), g_engfuncs.pfnGetInfoKeyBuffer(pEntity), "topcolor", "150");
	g_engfuncs.pfnSetClientKeyValue(pPlayer->entindex(), g_engfuncs.pfnGetInfoKeyBuffer(pEntity), "bottomcolor", "150");

	g_pGameRules->PlayerThink (pPlayer);
}


char *NameBot() { //Edit this function to come up with a name for your bot.
	return "Bot";
}

void BotConnect( edict_t *aEntity)
{
	edict_t *pEntity;
	char *botname;

	botname = NameBot();

	pEntity = g_engfuncs.pfnCreateFakeClient(botname);

	if (!pEntity) {
		ClientPrint(&aEntity->v, HUD_PRINTNOTIFY, UTIL_VarArgs( "Could not find any free client slots. Try setting maxplayers to a higher value.\n"));
		return;
	} //This means a fake client slot couldn't be created for some reason, so cancel the function.

	pEntity->v.netname = MAKE_STRING(botname); //Set the net name to the name we're using.

	BotClientPutInServer(pEntity);
}